"""
batterydataextractor.__init__

BatteryDataExtractor
"""

__version__ = "1.0.0"
__author__ = 'Shu Huang'
__credits__ = 'Molecular Engineering Group, Cavendish Laboratory'

from .doc.document import Document
